'use strict';

function _interopDefault (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }

var axios = _interopDefault(require('axios'));
var jsdom = _interopDefault(require('jsdom'));
var puppeteer = _interopDefault(require('puppeteer'));

/**
 * This create a result object and allows transform the given data returned.
 *
 * @param {object} result
 * @param {Function} fork
 * @return {object} Return the final result
 */
var createResult = (result, fork) => {
  return fork ? { ...result, ...fork(result) } : result
};

var getJSONContentFromChain = (selector, data) => {
  return selector
    .replace(/\[|\]\.?/g, ".")
    .split(".")
    .filter((s) => s)
    .reduce((acc, val) => acc && acc[val], data)
};

var getJSONContent = (jsonObject, selector) => {
  const { value, prefix = "", suffix = "" } = selector;
  const content = getJSONContentFromChain(value, jsonObject);

  if (content) {
    return `${prefix}${content}${suffix}`
  }

  return null
};

var getSourcesFromJSON = (list, keys) => {
  return list.map((jsonObject) => {
    let source = {};

    for (const key in keys) {
      if (Object.prototype.hasOwnProperty.call(keys, key)) {
        const selector = keys[key];
        const content = getJSONContent(jsonObject, selector);
        if (content) {
          source[key] = content;
        }
      }
    }

    return source
  })
};

/**
 * This extract a NodeList using a Query Selector item as root
 *
 * @param {string} html
 * @param {object} node
 * @param {object} options
 * @return {NodeList} // A list of HTMLElement items that match the given Query Selector
 */
var getNodeListFromSelector = (html, node, options = {}) => {
  const dom = new jsdom.JSDOM(html, options);
  const doc = dom.window.document;
  return doc.querySelectorAll(node.value)
};

/**
 * This helper format text when needed removing inconsistences
 *
 * @param {string} text
 * @param {boolean} parse
 * @return {string} Return text parsed/formatted
 */
var inlineTextFormat = (text, parse) => {
  if (!parse) {
    return text
  }

  let inlineValue = text.replace(/\n|\r/g, "");

  return inlineValue.replace(/\s+/g, " ").trim()
};

/**
 * This extract content from a HTMLElement node
 *
 * @param {HTMLElement} node
 * @param {object} selector
 * @return {string} Return content
 */
var getNodeContent = (node, selector) => {
  const { value, attribute, prefix = "", suffix = "", inline = true } = selector;
  const el = node.querySelector(value);
  if (attribute) {
    if (el) {
      return `${prefix}${inlineTextFormat(
        el.getAttribute(attribute),
        inline
      )}${suffix}`
    } else {
      return null
    }
  }

  if (el) {
    return `${prefix}${inlineTextFormat(el.textContent, inline)}${suffix}`
  }

  return null
};

/**
 * This extract sources from a given NodeList of HTMLElement items
 *
 * @param {NodeList} nodeList
 * @param {object} keys
 * @return {Array} Return a list of sources mapped
 */
var getSourcesFromNodeList = (nodeList, keys) => {
  let sources = [];

  nodeList.forEach((node) => {
    let source = {};

    for (const key in keys) {
      if (Object.prototype.hasOwnProperty.call(keys, key)) {
        const selector = keys[key];
        const content = getNodeContent(node, selector);

        if (content) {
          source[key] = content;
        }
      }
    }
    sources = [...sources, source];
  });

  return sources
};

/**
 * This auto scroll pages. It's really useful for any case, but mainly for lazy load pages.
 *
 * @param {PuppeteerPageClass} page https://github.com/puppeteer/puppeteer/blob/master/docs/api.md#class-pages
 */
const autoScroll = async (page) => {
  /* istanbul ignore next */ // TODO: Improve this test
  await page.evaluate(async () => {
    await new Promise((resolve) => {
      let totalHeight = 0;
      let distance = 100;
      let timer = setInterval(() => {
        let scrollHeight = document.body.scrollHeight;
        window.scrollBy(0, distance);
        totalHeight += distance;
        if (totalHeight >= scrollHeight) {
          clearInterval(timer);
          resolve();
        }
      }, 100);
    });
  });
};

/**
 * @param {string} url
 * @return {string} Return HTML content string
 */
var createSPARequest = async (url) => {
  const browser = await puppeteer.launch();
  const page = await browser.newPage();
  await page.goto(url);
  await autoScroll(page);
  const content = await page.content();
  await browser.close();
  return content
};

const getSourcesFromJSONObject = async ({
  options,
  selectors,
  metadata,
  fork,
}) => {
  try {
    const { root, ...rest } = selectors;
    const {
      request: { url, headers = {} },
      limit,
    } = options;
    const { data } = await axios({
      url,
      headers: {
        "user-agent": "node.js",
        ...headers,
      },
    });
    const list = root ? getJSONContentFromChain(root.value, data) : data;
    const sources = getSourcesFromJSON(list, rest);

    return createResult(
      {
        sources: sources.slice(0, limit),
        metadata,
      },
      fork
    )
  } catch ({ message }) {
    return createResult(
      {
        sources: [],
        metadata,
        error: message,
      },
      fork
    )
  }
};

var JSONEngine = {
  engine: "json",
  run: getSourcesFromJSONObject,
};

const getSourcesFromMarkup = async ({ options, selectors, metadata, fork }) => {
  try {
    const { root, ...rest } = selectors;
    const {
      request: { url, headers },
      xml = false,
      limit,
    } = options;
    const { data } = await axios({ url, headers });
    const nodeList = getNodeListFromSelector(data, root, {
      contentType: xml ? "text/xml" : "text/html",
    });
    const sources = getSourcesFromNodeList(nodeList, rest);

    return createResult(
      {
        sources: sources.slice(0, limit),
        metadata,
      },
      fork
    )
  } catch ({ message }) {
    return createResult(
      {
        sources: [],
        metadata,
        error: message,
      },
      fork
    )
  }
};

var markupEngine = {
  engine: "markup",
  run: getSourcesFromMarkup,
};

const getSourcesFromSPA = async ({ options, selectors, metadata, fork }) => {
  try {
    const { root, ...rest } = selectors;
    const {
      request: { url },
      limit,
    } = options;
    const data = await createSPARequest(url);
    const nodeList = getNodeListFromSelector(data, root);
    const sources = getSourcesFromNodeList(nodeList, rest);

    return createResult(
      {
        sources: sources.slice(0, limit),
        metadata,
      },
      fork
    )
  } catch ({ message }) {
    return createResult(
      {
        sources: [],
        metadata,
        error: message,
      },
      fork
    )
  }
};

var SPAEngine = {
  engine: "spa",
  run: getSourcesFromSPA,
};

/**
 * @param {Array<object>} schemas A list of schemas following the default and/or custom engines registered.
 * @param {Array<Promise>} customEngines A list of custom engines to be registered.
 * @return {Promise} Returns a promise with data sources.
 */
const Tatooine = (schemas, customEngines = []) => {
  const engines = [JSONEngine, markupEngine, SPAEngine, ...customEngines];
  let sources = [];

  schemas.map((schema) => {
    engines.forEach(({ engine, run }) => {
      if (engine === schema.engine) {
        sources = [...sources, run(schema)];
      }
    });
  });

  return Promise.all(sources)
};

module.exports = Tatooine;
//# sourceMappingURL=index.js.map
